var canvas, canvas2, activeCTX, ctx, ctx2, keyDown = false;

var coord = {
	prevX: 0,
	prevY: 0,
	currX: 0,
	currY: 0
};

var settings = {
	colour:'#000',
	thickness: 1,
	shape: 'free',
	width: window.innerWidth,
	height: window.innerHeight
};

var mouse = {
	flag: false,
	dot_flag: false,
	isDown: false
};

var strings = {
	ids: {
		canvas1: 'drawr-canvas',
		canvas2: 'drawr-canvas-store'
	},
	classes: {
		hasCanvas: 'has-drawr-canvas',
		screenshot: 'drawr-screenshot'
	},
	text: {
		clearCanvas: 'Want to clear the canvas?',
		screenshot: ' Screenshot using Drawr'
	}
};

function createCanvas(id) {
	var newCanvas = document.createElement('canvas');
	newCanvas.setAttribute('id', id);
	newCanvas.setAttribute('width', settings.width);
	newCanvas.setAttribute('height', settings.height);

	return newCanvas;
}

function removeCanvas() {
	var m = confirm(strings.text.clearCanvas);
	if (m) {
		document.documentElement.classList.remove(strings.classes.hasCanvas);
		document.getElementById(strings.ids.canvas1).remove();
		document.getElementById(strings.ids.canvas2).remove();
		document.removeEventListener("keydown", listenToTheKey);
		document.removeEventListener("keyup", keyUp);
	}
}

function cloneCanvas() {
	// quickest way to clear event bindings
	// create 2 news canvases and ctxs
	// reset variables and set activeCTX back to the top layer ctx
	var new_element = canvas.cloneNode(true);
	ctx = new_element.getContext('2d');
	ctx.drawImage(canvas, 0, 0);
	canvas.parentNode.replaceChild(new_element, canvas);
	canvas = new_element;

	var new_element2 = canvas2.cloneNode(true);
	ctx2 = new_element2.getContext('2d');
	ctx2.drawImage(canvas2, 0, 0);
	canvas2.parentNode.replaceChild(new_element2, canvas2);
	canvas2 = new_element2;

	activeCTX = ctx;
}

function storeCanvas() {
	// drawing finished, send canvas drawing to back canvas
	ctx2.drawImage(canvas, 0, 0);
	// clear top layer
	ctx.clearRect(0,0,settings.width,settings.height);
}

function cursorPosition(e) {
	coord.prevX = coord.currX;
	coord.prevY = coord.currY;
	coord.currX = e.clientX + (1.5*(e.clientX/100)); // horrible adjustment, pixel is drawn further negative offset to the left, the further right the cursor is
	coord.currY = e.clientY - canvas.offsetTop;
}

function mousedown(e) {
	if (!keyDown) {
		cursorPosition(e);
		mouse.flag = true;
		mouse.dot_flag = true;

		if (mouse.dot_flag) {
			activeCTX.beginPath();
			activeCTX.fillStyle = settings.colour;
			activeCTX.lineWidth = settings.thickness;
			activeCTX.lineCap = 'round';
			activeCTX.fillRect(coord.currX, coord.currY, 1, 1);
			mouse.dot_flag = false;
		}
	}
}

function mouseup() {
	mouse.flag = false;
	mouse.isDown = false;
	storeCanvas();
}

function mousemove(e) {
	if (mouse.flag) {
		cursorPosition(e);
		draw();
	}
}

function shapeMousemove(e) {
	if (!mouse.isDown) return;
	coord.currX = e.clientX - canvas.offsetLeft;
	coord.currY = e.clientY - canvas.offsetTop;
	if (settings.shape === 'circle') {
		drawCircle(coord.prevX, coord.prevY, coord.currX, coord.currY);
	} else if (settings.shape === 'rectangle') {
		drawRectangle();
	} else if (settings.shape === 'line') {
		draw();
	}
}

function shapeMousedown(e) {
	coord.prevX = e.clientX - canvas.offsetLeft;
	coord.prevY = e.clientY - canvas.offsetTop;
	mouse.isDown = true;
}

function listenToTheKey(e) {
	canvas.classList.remove('highlighted');
	keyDown = true;
	if (e.ctrlKey) {
		settings.shape = 'circle';
	} else if (e.shiftKey) {
		settings.shape = 'line';
	} else if (e.altKey) {
		settings.shape = 'rectangle';
	} else {
		settings.shape = 'free';
		keyDown = false;
	}

	attachBindings(settings.shape === 'free');
}

function keyUp(e) {
	keyDown = false;
	settings.shape = 'free';
	// dump live canvas data to the background canvas
	ctx2.drawImage(canvas, 0, 0);
	// no keys pressed to set 'free' to true to attach free draw bindings
	attachBindings(true);
}

function draw() {
	// if drawing a straight line, we need to keep clearing the last line
	if (settings.shape === 'line') {
		ctx.clearRect(0,0,settings.width,settings.height);
	}
	activeCTX.beginPath();
	activeCTX.moveTo(coord.prevX, coord.prevY);
	activeCTX.lineTo(coord.currX, coord.currY);
	activeCTX.strokeStyle = settings.colour;
	activeCTX.lineWidth = settings.thickness;
	activeCTX.lineCap = 'round';
	activeCTX.stroke();
}

function attachBindings(free) {
	cloneCanvas();
	if (!free) {
		canvas.addEventListener("mousemove", shapeMousemove, false);
		canvas.addEventListener("mouseup", mouseup, false);
		canvas.addEventListener("mousedown", shapeMousedown, false);
	} else {
		canvas.addEventListener("mousemove", mousemove, false);
		canvas.addEventListener("mousedown", mousedown, false);
		canvas.addEventListener("mouseup", mouseup, false);
	}
}

function drawRectangle() {
	// depending on the direction the user drags the rectangle
	// use either the the previous point or current point as the starting position
	var x = coord.currX > coord.prevX ? coord.prevX : coord.currX;
	var y = coord.currY > coord.prevY ? coord.prevY : coord.currY;
	var width = Math.abs(coord.currX - coord.prevX);
   	var height = Math.abs(coord.currY - coord.prevY);

   	ctx.clearRect(0, 0, settings.width, settings.height);
	ctx.beginPath();
    ctx.rect(x, y, width, height);
    ctx.fillStyle = "transparent";
    ctx.fill();
    ctx.lineWidth = settings.thickness;
    ctx.strokeStyle = settings.colour;
    ctx.stroke();
}

function drawCircle(x1, y1, x2, y2) {
	if (!mouse.isDown) return;
	ctx.clearRect(0, 0, settings.width, settings.height);
	var radiusX = (x2 - x1) * 0.5,   // radius for x based on input
		radiusY = (y2 - y1) * 0.5,   // radius for y based on input
		centerX = x1 + radiusX,      // calc center
		centerY = y1 + radiusY,
		step = 0.01,                 // resolution of ellipse
		a = step,                    // counter
		pi2 = Math.PI * 2 - step;    // end angle
	// start a new path
	ctx.beginPath();
	// set start point at angle 0
	ctx.moveTo(centerX + radiusX * Math.cos(0), centerY + radiusY * Math.sin(0));
	// create the ellipse    
	for(; a < pi2; a += step) {
		ctx.lineTo(centerX + radiusX * Math.cos(a), centerY + radiusY * Math.sin(a));
	}

	ctx.strokeStyle = settings.colour;
	ctx.lineWidth = settings.thickness;
	ctx.stroke();
}

function updateColour(newColour) {
	settings.colour = newColour;
	// if transparent
	if (settings.colour == 'rgba(0,0,0,1)') {
		// switch active ctx to the background one which has all drawings on it
		activeCTX = ctx2;
		// switch to 'remove' operation
		activeCTX.globalCompositeOperation = "destination-out";
		// make tool thick for easier erasing
		settings.thickness = 10;
	} else {
		// has a colour so user wants to draw, set operation to draw
		activeCTX.globalCompositeOperation = "source-over";
		// switch active ctx to top layer to draw on top
		activeCTX = ctx;
		// check for existing tool thickness, else default to 1px
		chrome.storage.local.get(['thickness'], function(result) {
			settings.thickness = result.thickness ? result.thickness : 1;
		});
	}
}

function updateThickness(newThickness) {
	settings.thickness = newThickness;
}

function takeScreenshot(image, title) {
	// create a link
	var downloadableImage = document.createElement('a');
	// set href to the screenshot url
	downloadableImage.setAttribute('href', image);
	// set the download attribute to desired title
	downloadableImage.setAttribute('download', title + strings.text.screenshot);
	// add class to the link to hide it
	downloadableImage.classList.add(strings.classes.screenshot);
	// add the link to the page
	document.body.appendChild(downloadableImage);
	// programmatically click it to trigger the automatic download of the screenshot
	downloadableImage.click();
}

function insertCanvas() {
	if (!document.getElementById(strings.ids.canvas1)) {
		canvas = createCanvas(strings.ids.canvas1);
		canvas.classList.add('highlighted');
		canvas.style.cursor = "pointer";
		ctx = canvas.getContext('2d');
		// activeCTX allows to toggle which canvas is being worked on for drawing and erasing
		activeCTX = ctx;
		
		// second canvas needed for erasing
		// once line or shape is drawn it is sent back to the second canvas
		// all news drawing happens on canvas/ctx and stored on canvas2/ctx2
		// erase removes from ctx2
		canvas2 = createCanvas(strings.ids.canvas2);
		ctx2 = canvas2.getContext('2d');

		// when canvas is first inserted, colour should be black
		chrome.storage.local.set({'colour': '#000'});

		chrome.storage.local.get(['thickness'], function(result) {
			settings.thickness = result.thickness ? result.thickness : 1;
		});

		document.documentElement.classList.add(strings.classes.hasCanvas);
		document.body.appendChild(canvas2);
		document.body.appendChild(canvas);
		// key bindings to determine between shapes or free draw
		document.addEventListener("keyup", keyUp, false);
		document.addEventListener("keydown", listenToTheKey, false);
		// pass in true for free drawing
		attachBindings(true);
	}
}

function storageChangeSwitch(key, changes) {
	// run functions based on user selection in popup
	var item = changes[key].newValue;
	switch(key) {
		case 'colour':
			updateColour(item);
			break;
		case 'thickness':
			updateThickness(item);
			break;
		case 'action':
			if (item === 'draw' || item === 'triggered') {
				insertCanvas();
			} else if (item === 'clear') {
				removeCanvas();
			} else if (item === 'screenshot') {
				var title = changes.screenshotTitle ? changes.screenshotTitle.newValue.title : "";
				takeScreenshot(changes.screenshotURL.newValue, title);
			}
			break;
	}
}

chrome.storage.onChanged.addListener(function(changes) {
	// popup interaction updates storage
	// this listens for changes
	for (key in changes) {
		storageChangeSwitch(key, changes);
	}
});